cimport pysndlib.csndlib as csndlib

#==================================================================================
# The code is part of an attempt at translation of Bill Schottstedaet's clm 
# available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

cdef extern from "clm.h":
    """
    typedef mus_float_t (*connect_points_cb)(mus_float_t val);
    typedef mus_float_t (*input_cb)(void *arg, int direction);
    typedef int (*edit_cb)(void *closure);
    typedef bool (*analyze_cb)(void *arg, mus_float_t (*input)(void *arg1, int direction));
	typedef mus_float_t (*synthesize_cb)(void *arg);
	typedef void (*detour_cb)(mus_any *ptr, mus_long_t val);

    typedef mus_any* mus_any_ptr;
    typedef mus_any** mus_any_ptr_ptr;
    """
    
    
    ctypedef signed long int64_t
    ctypedef double mus_float_t
    ctypedef unsigned long uint64_t
    ctypedef int64_t mus_long_t 
    
    ctypedef struct mus_any_class:
        pass
        
    ctypedef struct mus_any:
        pass

    ctypedef enum mus_clm_extended_t:
        pass
        
    ctypedef enum mus_interp_t:
        pass
    
    ctypedef enum mus_fft_window_t:
        pass
    
    ctypedef enum mus_spectrum_t:
        pass
         
    ctypedef enum mus_polynomial_t:
        pass
        
        
    ctypedef enum mus_sample_t:
        pass
        
#     ctypedef enum mus_header_t:
#         pass
        
    ctypedef mus_any* mus_any_ptr
    
    ctypedef mus_any** mus_any_ptr_ptr
    
    ctypedef mus_float_t (*connect_points_cb)(mus_float_t val)
    ctypedef mus_float_t (*input_cb)(void *arg, int direction)
    ctypedef bint (*analyze_cb)(void *arg, mus_float_t (*input)(void *arg1, int direction))
    ctypedef int (*edit_cb)(void *closure)
    ctypedef void (*detour_cb)(mus_any *ptr, mus_long_t val)
    ctypedef mus_float_t (*synthesize_cb)(void *arg)

    void mus_initialize()

    int mus_make_generator_type()
    


    mus_any_class *mus_generator_class(mus_any *ptr)
    mus_any_class *mus_make_generator(int type, char *name, 
                         void (*release)(mus_any *ptr), 
                         char *(*describe)(mus_any *ptr), 
                         bint (*equalp)(mus_any *gen1, mus_any *gen2))

    void mus_generator_set_length(mus_any_class *p, mus_long_t (*length)(mus_any *ptr))
    void mus_generator_set_scaler(mus_any_class *p, mus_float_t (*scaler)(mus_any *ptr))
    void mus_generator_set_channels(mus_any_class *p, int (*channels)(mus_any *ptr))
    void mus_generator_set_location(mus_any_class *p, mus_long_t (*location)(mus_any *ptr))
    void mus_generator_set_set_location(mus_any_class *p, mus_long_t (*set_location)(mus_any *ptr, mus_long_t loc))
    void mus_generator_set_channel(mus_any_class *p, int (*channel)(mus_any *ptr))
    void mus_generator_set_file_name(mus_any_class *p, char *(*file_name)(mus_any *ptr))
    void mus_generator_set_extended_type(mus_any_class *p, mus_clm_extended_t extended_type)
    void mus_generator_set_read_sample(mus_any_class *p, mus_float_t (*read_sample)(mus_any *ptr, mus_long_t samp, int chan))
    void mus_generator_set_feeders(mus_any *g, 
                      mus_float_t (*feed)(void *arg, int direction),
                      mus_float_t (*block_feed)(void *arg, int direction, mus_float_t *block, mus_long_t start, mus_long_t end))
    void mus_generator_copy_feeders(mus_any *dest, mus_any *source)

    mus_float_t mus_radians_to_hz(mus_float_t radians)
    mus_float_t mus_hz_to_radians(mus_float_t hz)
    mus_float_t mus_degrees_to_radians(mus_float_t degrees)
    mus_float_t mus_radians_to_degrees(mus_float_t radians)
    mus_float_t mus_db_to_linear(mus_float_t x)
    mus_float_t mus_linear_to_db(mus_float_t x)
    mus_float_t mus_odd_multiple(mus_float_t x, mus_float_t y)
    mus_float_t mus_even_multiple(mus_float_t x, mus_float_t y)
    mus_float_t mus_odd_weight(mus_float_t x)
    mus_float_t mus_even_weight(mus_float_t x)
    const char *mus_interp_type_to_string(int type)

    mus_float_t mus_srate()
    mus_float_t mus_set_srate(mus_float_t val)
    mus_long_t mus_seconds_to_samples(mus_float_t secs)
    mus_float_t mus_samples_to_seconds(mus_long_t samps)
    int mus_array_print_length()
    int mus_set_array_print_length(int val)
    mus_float_t mus_float_equal_fudge_factor()
    mus_float_t mus_set_float_equal_fudge_factor(mus_float_t val)

    mus_float_t mus_ring_modulate(mus_float_t s1, mus_float_t s2)
    mus_float_t mus_amplitude_modulate(mus_float_t s1, mus_float_t s2, mus_float_t s3)
    mus_float_t mus_contrast_enhancement(mus_float_t sig, mus_float_t index)
    mus_float_t mus_dot_product(mus_float_t *data1, mus_float_t *data2, mus_long_t size)
#     #if HAVE_COMPLEX_TRIG
#     complex double mus_edot_product(complex double freq, complex double *data, mus_long_t size)
#     #endif

    bint mus_arrays_are_equal(mus_float_t *arr1, mus_float_t *arr2, mus_float_t fudge, mus_long_t len)
    mus_float_t mus_polynomial(mus_float_t *coeffs, mus_float_t x, int ncoeffs)
    void mus_rectangular_to_polar(mus_float_t *rl, mus_float_t *im, mus_long_t size)
    void mus_rectangular_to_magnitudes(mus_float_t *rl, mus_float_t *im, mus_long_t size)
    void mus_polar_to_rectangular(mus_float_t *rl, mus_float_t *im, mus_long_t size)
    mus_float_t mus_array_interp(mus_float_t *wave, mus_float_t phase, mus_long_t size)
    mus_float_t mus_bessi0(mus_float_t x)
    mus_float_t mus_interpolate(mus_interp_t type, mus_float_t x, mus_float_t *table, mus_long_t table_size, mus_float_t y)
    bint mus_is_interp_type(int val)
    bint mus_is_fft_window(int val)

    int mus_sample_type_zero(mus_sample_t samp_type)
    mus_float_t (*mus_run_function(mus_any *g))(mus_any *gen, mus_float_t arg1, mus_float_t arg2)
    mus_float_t (*mus_run1_function(mus_any *g))(mus_any *gen, mus_float_t arg)


#     /* -------- generic functions -------- */

    int mus_type(mus_any *ptr)
    void mus_free(mus_any *ptr)
    char *mus_describe(mus_any *gen)
    bint mus_equalp(mus_any *g1, mus_any *g2)
    mus_float_t mus_phase(mus_any *gen)
    mus_float_t mus_set_phase(mus_any *gen, mus_float_t val)
    mus_float_t mus_set_frequency(mus_any *gen, mus_float_t val)
    mus_float_t mus_frequency(mus_any *gen)
    mus_float_t mus_run(mus_any *gen, mus_float_t arg1, mus_float_t arg2)
    mus_long_t mus_length(mus_any *gen)
    mus_long_t mus_set_length(mus_any *gen, mus_long_t len)
    mus_long_t mus_order(mus_any *gen)
    mus_float_t *mus_data(mus_any *gen)
    mus_float_t *mus_set_data(mus_any *gen, mus_float_t *data)
    const char *mus_name(mus_any *ptr)
    mus_float_t mus_scaler(mus_any *gen)
    mus_float_t mus_set_scaler(mus_any *gen, mus_float_t val)
    mus_float_t mus_offset(mus_any *gen)
    mus_float_t mus_set_offset(mus_any *gen, mus_float_t val)
    mus_float_t mus_width(mus_any *gen)
    mus_float_t mus_set_width(mus_any *gen, mus_float_t val)
    char *mus_file_name(mus_any *ptr)
    void mus_reset(mus_any *ptr)
    mus_any *mus_copy(mus_any *gen)
    mus_float_t *mus_xcoeffs(mus_any *ptr)
    mus_float_t *mus_ycoeffs(mus_any *ptr)
    mus_float_t mus_xcoeff(mus_any *ptr, int index)
    mus_float_t mus_set_xcoeff(mus_any *ptr, int index, mus_float_t val)
    mus_float_t mus_ycoeff(mus_any *ptr, int index)
    mus_float_t mus_set_ycoeff(mus_any *ptr, int index, mus_float_t val)
    mus_float_t mus_increment(mus_any *rd)
    mus_float_t mus_set_increment(mus_any *rd, mus_float_t dir)
    mus_long_t mus_location(mus_any *rd)
    mus_long_t mus_set_location(mus_any *rd, mus_long_t loc)
    int mus_channel(mus_any *rd)
    int mus_channels(mus_any *ptr)
    int mus_position(mus_any *ptr) #/* only C, envs (snd-env.c), shares slot with mus_channels */
    int mus_interp_type(mus_any *ptr)
    mus_long_t mus_ramp(mus_any *ptr)
    mus_long_t mus_set_ramp(mus_any *ptr, mus_long_t val)
    mus_long_t mus_hop(mus_any *ptr)
    mus_long_t mus_set_hop(mus_any *ptr, mus_long_t val)
    mus_float_t mus_feedforward(mus_any *gen)
    mus_float_t mus_set_feedforward(mus_any *gen, mus_float_t val)
    mus_float_t mus_feedback(mus_any *rd)
    mus_float_t mus_set_feedback(mus_any *rd, mus_float_t dir)

    bint mus_phase_exists(mus_any *gen)
    bint mus_frequency_exists(mus_any *gen)
    bint mus_length_exists(mus_any *gen)
    bint mus_order_exists(mus_any *gen)
    bint mus_data_exists(mus_any *gen)
    bint mus_name_exists(mus_any *gen)
    bint mus_scaler_exists(mus_any *gen)
    bint mus_offset_exists(mus_any *gen)
    bint mus_width_exists(mus_any *gen)
    bint mus_file_name_exists(mus_any *gen)
    bint mus_xcoeffs_exists(mus_any *gen)
    bint mus_ycoeffs_exists(mus_any *gen)
    bint mus_increment_exists(mus_any *gen)
    bint mus_location_exists(mus_any *gen)
    bint mus_channel_exists(mus_any *gen)
    bint mus_channels_exists(mus_any *gen)
    bint mus_position_exists(mus_any *gen)
    bint mus_interp_type_exists(mus_any *gen)
    bint mus_ramp_exists(mus_any *gen)
    bint mus_hop_exists(mus_any *gen)
    bint mus_feedforward_exists(mus_any *gen)
    bint mus_feedback_exists(mus_any *gen)


#     /* -------- generators -------- */

    mus_float_t mus_oscil(mus_any *o, mus_float_t fm, mus_float_t pm)
    mus_float_t mus_oscil_unmodulated(mus_any *ptr)
    mus_float_t mus_oscil_fm(mus_any *ptr, mus_float_t fm)
    mus_float_t mus_oscil_pm(mus_any *ptr, mus_float_t pm)
    bint mus_is_oscil(mus_any *ptr)
    mus_any *mus_make_oscil(mus_float_t freq, mus_float_t phase)

    bint mus_is_oscil_bank(mus_any *ptr)
    mus_float_t mus_oscil_bank(mus_any *ptr)
    mus_any *mus_make_oscil_bank(int size, mus_float_t *freqs, mus_float_t *phases, mus_float_t *amps, bint stable)

    mus_any *mus_make_ncos(mus_float_t freq, int n)
    mus_float_t mus_ncos(mus_any *ptr, mus_float_t fm)
    bint mus_is_ncos(mus_any *ptr)

    mus_any *mus_make_nsin(mus_float_t freq, int n)
    mus_float_t mus_nsin(mus_any *ptr, mus_float_t fm)
    bint mus_is_nsin(mus_any *ptr)

    mus_any *mus_make_nrxysin(mus_float_t frequency, mus_float_t y_over_x, int n, mus_float_t r)
    mus_float_t mus_nrxysin(mus_any *ptr, mus_float_t fm)
    bint mus_is_nrxysin(mus_any *ptr)

    mus_any *mus_make_nrxycos(mus_float_t frequency, mus_float_t y_over_x, int n, mus_float_t r)
    mus_float_t mus_nrxycos(mus_any *ptr, mus_float_t fm)
    bint mus_is_nrxycos(mus_any *ptr)

    mus_any *mus_make_rxykcos(mus_float_t freq, mus_float_t phase, mus_float_t r, mus_float_t ratio)
    mus_float_t mus_rxykcos(mus_any *ptr, mus_float_t fm)
    bint mus_is_rxykcos(mus_any *ptr)

    mus_any *mus_make_rxyksin(mus_float_t freq, mus_float_t phase, mus_float_t r, mus_float_t ratio)
    mus_float_t mus_rxyksin(mus_any *ptr, mus_float_t fm)
    bint mus_is_rxyksin(mus_any *ptr)

    mus_float_t mus_delay(mus_any *gen, mus_float_t input, mus_float_t pm)
    mus_float_t mus_delay_unmodulated(mus_any *ptr, mus_float_t input)
    mus_float_t mus_tap(mus_any *gen, mus_float_t loc)
    mus_float_t mus_tap_unmodulated(mus_any *gen)
    mus_any *mus_make_delay(int size, mus_float_t *line, int line_size, mus_interp_t type)
    bint mus_is_delay(mus_any *ptr)
    bint mus_is_tap(mus_any *ptr)
    mus_float_t mus_delay_tick(mus_any *ptr, mus_float_t input)
    mus_float_t mus_delay_unmodulated_noz(mus_any *ptr, mus_float_t input)

    mus_float_t mus_comb(mus_any *gen, mus_float_t input, mus_float_t pm)
    mus_float_t mus_comb_unmodulated(mus_any *gen, mus_float_t input)
    mus_any *mus_make_comb(mus_float_t scaler, int size, mus_float_t *line, int line_size, mus_interp_t type)
    bint mus_is_comb(mus_any *ptr)
    mus_float_t mus_comb_unmodulated_noz(mus_any *ptr, mus_float_t input)

    mus_float_t mus_comb_bank(mus_any *bank, mus_float_t inval)
    mus_any *mus_make_comb_bank(int size, mus_any **combs)
    bint mus_is_comb_bank(mus_any *g)

    mus_float_t mus_notch(mus_any *gen, mus_float_t input, mus_float_t pm)
    mus_float_t mus_notch_unmodulated(mus_any *gen, mus_float_t input)
    mus_any *mus_make_notch(mus_float_t scaler, int size, mus_float_t *line, int line_size, mus_interp_t type)
    bint mus_is_notch(mus_any *ptr)

    mus_float_t mus_all_pass(mus_any *gen, mus_float_t input, mus_float_t pm)
    mus_float_t mus_all_pass_unmodulated(mus_any *gen, mus_float_t input)
    mus_any *mus_make_all_pass(mus_float_t backward, mus_float_t forward, int size, mus_float_t *line, int line_size, mus_interp_t type)
    bint mus_is_all_pass(mus_any *ptr)
    mus_float_t mus_all_pass_unmodulated_noz(mus_any *ptr, mus_float_t input)

    mus_float_t mus_all_pass_bank(mus_any *bank, mus_float_t inval)
    mus_any *mus_make_all_pass_bank(int size, mus_any **combs)
    bint mus_is_all_pass_bank(mus_any *g)

    mus_any *mus_make_moving_average(int size, mus_float_t *line)
    mus_any *mus_make_moving_average_with_initial_sum(int size, mus_float_t *line, mus_float_t sum)
    bint mus_is_moving_average(mus_any *ptr)
    mus_float_t mus_moving_average(mus_any *ptr, mus_float_t input)

    mus_any *mus_make_moving_max(int size, mus_float_t *line)
    bint mus_is_moving_max(mus_any *ptr)
    mus_float_t mus_moving_max(mus_any *ptr, mus_float_t input)

    mus_any *mus_make_moving_norm(int size, mus_float_t *line, mus_float_t norm)
    bint mus_is_moving_norm(mus_any *ptr)
    mus_float_t mus_moving_norm(mus_any *ptr, mus_float_t input)

    mus_float_t mus_table_lookup(mus_any *gen, mus_float_t fm)
    mus_float_t mus_table_lookup_unmodulated(mus_any *gen)
    mus_any *mus_make_table_lookup(mus_float_t freq, mus_float_t phase, mus_float_t *wave, mus_long_t wave_size, mus_interp_t type)
    bint mus_is_table_lookup(mus_any *ptr)
    mus_float_t *mus_partials_to_wave(mus_float_t *partial_data, int partials, mus_float_t *table, mus_long_t table_size, bint normalize)
    mus_float_t *mus_phase_partials_to_wave(mus_float_t *partial_data, int partials, mus_float_t *table, mus_long_t table_size, bint normalize)

    mus_float_t mus_sawtooth_wave(mus_any *gen, mus_float_t fm)
    mus_any *mus_make_sawtooth_wave(mus_float_t freq, mus_float_t amp, mus_float_t phase)
    bint mus_is_sawtooth_wave(mus_any *gen)

    mus_float_t mus_square_wave(mus_any *gen, mus_float_t fm)
    mus_any *mus_make_square_wave(mus_float_t freq, mus_float_t amp, mus_float_t phase)
    bint mus_is_square_wave(mus_any *gen)

    mus_float_t mus_triangle_wave(mus_any *gen, mus_float_t fm)
    mus_any *mus_make_triangle_wave(mus_float_t freq, mus_float_t amp, mus_float_t phase)
    bint mus_is_triangle_wave(mus_any *gen)
    mus_float_t mus_triangle_wave_unmodulated(mus_any *ptr)

    mus_float_t mus_pulse_train(mus_any *gen, mus_float_t fm)
    mus_any *mus_make_pulse_train(mus_float_t freq, mus_float_t amp, mus_float_t phase)
    bint mus_is_pulse_train(mus_any *gen)
    mus_float_t mus_pulse_train_unmodulated(mus_any *ptr)

    void mus_set_rand_seed(uint64_t seed)
    uint64_t mus_rand_seed()
    mus_float_t mus_random(mus_float_t amp)
    mus_float_t mus_frandom(mus_float_t amp)
    int mus_irandom(int amp)

    mus_float_t mus_rand(mus_any *gen, mus_float_t fm)
    mus_any *mus_make_rand(mus_float_t freq, mus_float_t base)
    bint mus_is_rand(mus_any *ptr)
    mus_any *mus_make_rand_with_distribution(mus_float_t freq, mus_float_t base, mus_float_t *distribution, int distribution_size)


    mus_float_t mus_rand_interp(mus_any *gen, mus_float_t fm)
    mus_any *mus_make_rand_interp(mus_float_t freq, mus_float_t base)
    bint mus_is_rand_interp(mus_any *ptr)
    mus_any *mus_make_rand_interp_with_distribution(mus_float_t freq, mus_float_t base, mus_float_t *distribution, int distribution_size)
    mus_float_t mus_rand_interp_unmodulated(mus_any *ptr)
    mus_float_t mus_rand_unmodulated(mus_any *ptr)

    mus_float_t mus_asymmetric_fm(mus_any *gen, mus_float_t index, mus_float_t fm)
    mus_float_t mus_asymmetric_fm_unmodulated(mus_any *gen, mus_float_t index)
    mus_any *mus_make_asymmetric_fm(mus_float_t freq, mus_float_t phase, mus_float_t r, mus_float_t ratio)
    bint mus_is_asymmetric_fm(mus_any *ptr)

    mus_float_t mus_one_zero(mus_any *gen, mus_float_t input)
    mus_any *mus_make_one_zero(mus_float_t a0, mus_float_t a1)
    bint mus_is_one_zero(mus_any *gen)

    mus_float_t mus_one_pole(mus_any *gen, mus_float_t input)
    mus_any *mus_make_one_pole(mus_float_t a0, mus_float_t b1)
    bint mus_is_one_pole(mus_any *gen)

    mus_float_t mus_two_zero(mus_any *gen, mus_float_t input)
    mus_any *mus_make_two_zero(mus_float_t a0, mus_float_t a1, mus_float_t a2)
    bint mus_is_two_zero(mus_any *gen)
    mus_any *mus_make_two_zero_from_frequency_and_radius(mus_float_t frequency, mus_float_t radius)

    mus_float_t mus_two_pole(mus_any *gen, mus_float_t input)
    mus_any *mus_make_two_pole(mus_float_t a0, mus_float_t b1, mus_float_t b2)
    bint mus_is_two_pole(mus_any *gen)
    mus_any *mus_make_two_pole_from_frequency_and_radius(mus_float_t frequency, mus_float_t radius)

    mus_float_t mus_one_pole_all_pass(mus_any *f, mus_float_t input)
    mus_any *mus_make_one_pole_all_pass(int size, mus_float_t coeff)
    bint mus_is_one_pole_all_pass(mus_any *ptr)

    mus_float_t mus_formant(mus_any *ptr, mus_float_t input) 
    mus_any *mus_make_formant(mus_float_t frequency, mus_float_t radius)
    bint mus_is_formant(mus_any *ptr)
    mus_float_t mus_set_formant_frequency(mus_any *ptr, mus_float_t freq_in_hz)
    void mus_set_formant_radius_and_frequency(mus_any *ptr, mus_float_t radius, mus_float_t frequency)
    mus_float_t mus_formant_with_frequency(mus_any *ptr, mus_float_t input, mus_float_t freq_in_radians)

    mus_float_t mus_formant_bank(mus_any *bank, mus_float_t inval)
    mus_float_t mus_formant_bank_with_inputs(mus_any *bank, mus_float_t *inval)
    mus_any *mus_make_formant_bank(int size, mus_any **formants, mus_float_t *amps)
    bint mus_is_formant_bank(mus_any *g)

    mus_float_t mus_firmant(mus_any *ptr, mus_float_t input)
    mus_any *mus_make_firmant(mus_float_t frequency, mus_float_t radius)
    bint mus_is_firmant(mus_any *ptr)
    mus_float_t mus_firmant_with_frequency(mus_any *ptr, mus_float_t input, mus_float_t freq_in_radians)

    mus_float_t mus_filter(mus_any *ptr, mus_float_t input)
    mus_any *mus_make_filter(int order, mus_float_t *xcoeffs, mus_float_t *ycoeffs, mus_float_t *state)
    bint mus_is_filter(mus_any *ptr)

    mus_float_t mus_fir_filter(mus_any *ptr, mus_float_t input)
    mus_any *mus_make_fir_filter(int order, mus_float_t *xcoeffs, mus_float_t *state)
    bint mus_is_fir_filter(mus_any *ptr)

    mus_float_t mus_iir_filter(mus_any *ptr, mus_float_t input)
    mus_any *mus_make_iir_filter(int order, mus_float_t *ycoeffs, mus_float_t *state)
    bint mus_is_iir_filter(mus_any *ptr)
    mus_float_t *mus_make_fir_coeffs(int order, mus_float_t *env, mus_float_t *aa)

    mus_float_t *mus_filter_set_xcoeffs(mus_any *ptr, mus_float_t *new_data)
    mus_float_t *mus_filter_set_ycoeffs(mus_any *ptr, mus_float_t *new_data)
    int mus_filter_set_order(mus_any *ptr, int order)

    mus_float_t mus_filtered_comb(mus_any *ptr, mus_float_t input, mus_float_t pm)
    mus_float_t mus_filtered_comb_unmodulated(mus_any *ptr, mus_float_t input)
    bint mus_is_filtered_comb(mus_any *ptr)
    mus_any *mus_make_filtered_comb(mus_float_t scaler, int size, mus_float_t *line, int line_size, mus_interp_t type, mus_any *filt)

    mus_float_t mus_filtered_comb_bank(mus_any *bank, mus_float_t inval)
    mus_any *mus_make_filtered_comb_bank(int size, mus_any **combs)
    bint mus_is_filtered_comb_bank(mus_any *g)

    mus_float_t mus_wave_train(mus_any *gen, mus_float_t fm)
    mus_float_t mus_wave_train_unmodulated(mus_any *gen)
    mus_any *mus_make_wave_train(mus_float_t freq, mus_float_t phase, mus_float_t *wave, mus_long_t wsize, mus_interp_t type)
    bint mus_is_wave_train(mus_any *gen)

    mus_float_t *mus_partials_to_polynomial(int npartials, mus_float_t *partials, mus_polynomial_t kind)
    mus_float_t *mus_normalize_partials(int num_partials, mus_float_t *partials)

    mus_any *mus_make_polyshape(mus_float_t frequency, mus_float_t phase, mus_float_t *coeffs, int size, int cheby_choice)
    mus_float_t mus_polyshape(mus_any *ptr, mus_float_t index, mus_float_t fm)
    mus_float_t mus_polyshape_unmodulated(mus_any *ptr, mus_float_t index)
    #define mus_polyshape_no_input(Obj) mus_polyshape(Obj, 1.0, 0.0)
    bint mus_is_polyshape(mus_any *ptr)

    mus_any *mus_make_polywave(mus_float_t frequency, mus_float_t *coeffs, int n, int cheby_choice)
    mus_any *mus_make_polywave_tu(mus_float_t frequency, mus_float_t *tcoeffs, mus_float_t *ucoeffs, int n)
    bint mus_is_polywave(mus_any *ptr)
    mus_float_t mus_polywave_unmodulated(mus_any *ptr)
    mus_float_t mus_polywave(mus_any *ptr, mus_float_t fm)
    mus_float_t mus_chebyshev_t_sum(mus_float_t x, int n, mus_float_t *tn)
    mus_float_t mus_chebyshev_u_sum(mus_float_t x, int n, mus_float_t *un)
    mus_float_t mus_chebyshev_tu_sum(mus_float_t x, int n, mus_float_t *tn, mus_float_t *un)
    mus_float_t (*mus_polywave_function(mus_any *g))(mus_any *gen, mus_float_t fm)

    mus_float_t mus_env(mus_any *ptr)
    mus_any *mus_make_env(mus_float_t *brkpts, int npts, mus_float_t scaler, mus_float_t offset, mus_float_t base, 
                 mus_float_t duration, mus_long_t end, mus_float_t *odata)
    bint mus_is_env(mus_any *ptr)
    mus_float_t mus_env_interp(mus_float_t x, mus_any *env)
    mus_long_t *mus_env_passes(mus_any *gen)        #/* for Snd */
    mus_float_t *mus_env_rates(mus_any *gen)        #/* for Snd */
    mus_float_t mus_env_offset(mus_any *gen)        #/* for Snd */
    mus_float_t mus_env_scaler(mus_any *gen)        #/* for Snd */
    mus_float_t mus_env_initial_power(mus_any *gen) #/* for Snd */
    int mus_env_breakpoints(mus_any *gen)     # /* for Snd */
    mus_float_t mus_env_any(mus_any *e, mus_float_t (*connect_points)(mus_float_t val))
    mus_float_t (*mus_env_function(mus_any *g))(mus_any *gen)


    mus_any *mus_make_pulsed_env(mus_any *e, mus_any *p)
    bint mus_is_pulsed_env(mus_any *ptr)
    mus_float_t mus_pulsed_env(mus_any *pl, mus_float_t inval)
    mus_float_t mus_pulsed_env_unmodulated(mus_any *pl)

    bint mus_is_file_to_sample(mus_any *ptr)
    mus_any *mus_make_file_to_sample(const char *filename)
    mus_any *mus_make_file_to_sample_with_buffer_size(const char *filename, mus_long_t buffer_size)
    mus_float_t mus_file_to_sample(mus_any *ptr, mus_long_t samp, int chan)

    mus_float_t mus_readin(mus_any *rd)
    mus_any *mus_make_readin_with_buffer_size(const char *filename, int chan, mus_long_t start, int direction, mus_long_t buffer_size)
    #define mus_make_readin(Filename, Chan, Start, Direction) mus_make_readin_with_buffer_size(Filename, Chan, Start, Direction, mus_file_buffer_size())
    bint mus_is_readin(mus_any *ptr)

    bint mus_is_output(mus_any *ptr)
    bint mus_is_input(mus_any *ptr)
    mus_float_t mus_in_any(mus_long_t frample, int chan, mus_any *IO)
    bint mus_in_any_is_safe(mus_any *IO)

    #/* new 6.0 */
    mus_float_t *mus_file_to_frample(mus_any *ptr, mus_long_t samp, mus_float_t *f)
    mus_any *mus_make_file_to_frample(const char *filename)
    bint mus_is_file_to_frample(mus_any *ptr)
    mus_any *mus_make_file_to_frample_with_buffer_size(const char *filename, mus_long_t buffer_size)
    mus_float_t *mus_frample_to_frample(mus_float_t *matrix, int mx_chans, mus_float_t *in_samps, int in_chans, mus_float_t *out_samps, int out_chans)

    bint mus_is_frample_to_file(mus_any *ptr)
    mus_float_t *mus_frample_to_file(mus_any *ptr, mus_long_t samp, mus_float_t *data)
    mus_any *mus_make_frample_to_file_with_comment(const char *filename, int chans, mus_sample_t samp_type, csndlib.mus_header_t head_type, const char *comment)
    #define mus_make_frample_to_file(Filename, Chans, SampType, HeadType) mus_make_frample_to_file_with_comment(Filename, Chans, SampType, HeadType, NULL)
    mus_any *mus_continue_frample_to_file(const char *filename)

    void mus_file_mix_with_reader_and_writer(mus_any *outf, mus_any *inf,
                            mus_long_t out_start, mus_long_t out_framples, mus_long_t in_start, 
                            mus_float_t *mx, int mx_chans, mus_any ***envs)
    void mus_file_mix(const char *outfile, const char *infile, 
                 mus_long_t out_start, mus_long_t out_framples, mus_long_t in_start, 
                 mus_float_t *mx, int mx_chans, mus_any ***envs)

    bint mus_is_sample_to_file(mus_any *ptr)
    mus_any *mus_make_sample_to_file_with_comment(const char *filename, int out_chans, mus_sample_t samp_type, csndlib.mus_header_t head_type, const char *comment)
    #define mus_make_sample_to_file(Filename, Chans, SampType, HeadType) mus_make_sample_to_file_with_comment(Filename, Chans, SampType, HeadType, NULL)
    mus_float_t mus_sample_to_file(mus_any *ptr, mus_long_t samp, int chan, mus_float_t val)
    mus_any *mus_continue_sample_to_file(const char *filename)
    int mus_close_file(mus_any *ptr)
    mus_any *mus_sample_to_file_add(mus_any *out1, mus_any *out2)

    mus_float_t mus_out_any(mus_long_t frample, mus_float_t val, int chan, mus_any *IO)
    mus_float_t mus_safe_out_any_to_file(mus_long_t samp, mus_float_t val, int chan, mus_any *IO)
    bint mus_out_any_is_safe(mus_any *IO)
    mus_float_t mus_out_any_to_file(mus_any *ptr, mus_long_t samp, int chan, mus_float_t val)

    void mus_locsig(mus_any *ptr, mus_long_t loc, mus_float_t val)
    mus_any *mus_make_locsig(mus_float_t degree, mus_float_t distance, mus_float_t reverb, int chans, 
                    mus_any *output, int rev_chans, mus_any *revput, mus_interp_t type)
    bint mus_is_locsig(mus_any *ptr)
    mus_float_t mus_locsig_ref(mus_any *ptr, int chan)
    mus_float_t mus_locsig_set(mus_any *ptr, int chan, mus_float_t val)
    mus_float_t mus_locsig_reverb_ref(mus_any *ptr, int chan)
    mus_float_t mus_locsig_reverb_set(mus_any *ptr, int chan, mus_float_t val)
    void mus_move_locsig(mus_any *ptr, mus_float_t degree, mus_float_t distance)
    mus_float_t *mus_locsig_outf(mus_any *ptr)
    mus_float_t *mus_locsig_revf(mus_any *ptr)
    void *mus_locsig_closure(mus_any *ptr)
    void mus_locsig_set_detour(mus_any *ptr, void (*detour)(mus_any *ptr, mus_long_t val))
    int mus_locsig_channels(mus_any *ptr)
    int mus_locsig_reverb_channels(mus_any *ptr)

    bint mus_is_move_sound(mus_any *ptr)
    mus_float_t mus_move_sound(mus_any *ptr, mus_long_t loc, mus_float_t val)
    mus_any *mus_make_move_sound(mus_long_t start, mus_long_t end, int out_channels, int rev_channels,
                    mus_any *doppler_delay, mus_any *doppler_env, mus_any *rev_env,
                    mus_any **out_delays, mus_any **out_envs, mus_any **rev_envs,
                    int *out_map, mus_any *output, mus_any *revput, bint free_arrays, bint free_gens)
    mus_float_t *mus_move_sound_outf(mus_any *ptr)
    mus_float_t *mus_move_sound_revf(mus_any *ptr)
    void *mus_move_sound_closure(mus_any *ptr)
    void mus_move_sound_set_detour(mus_any *ptr, void (*detour)(mus_any *ptr, mus_long_t val))
    int mus_move_sound_channels(mus_any *ptr)
    int mus_move_sound_reverb_channels(mus_any *ptr)

    mus_any *mus_make_src(mus_float_t (*input)(void *arg, int direction), mus_float_t srate, int width, void *closure)
    mus_any *mus_make_src_with_init(mus_float_t (*input)(void *arg, int direction), mus_float_t srate, int width, void *closure, void (*init)(void *p, mus_any *g))
    mus_float_t mus_src(mus_any *srptr, mus_float_t sr_change, mus_float_t (*input)(void *arg, int direction))
    bint mus_is_src(mus_any *ptr)
    mus_float_t *mus_src_20(mus_any *srptr, mus_float_t *in_data, mus_long_t dur)
    mus_float_t *mus_src_05(mus_any *srptr, mus_float_t *in_data, mus_long_t dur)
    void mus_src_to_buffer(mus_any *srptr, mus_float_t (*input)(void *arg, int direction), mus_float_t *out_data, mus_long_t dur)
    void mus_src_init(mus_any *ptr)

    bint mus_is_convolve(mus_any *ptr)
    mus_float_t mus_convolve(mus_any *ptr, mus_float_t (*input)(void *arg, int direction))
    mus_any *mus_make_convolve(mus_float_t (*input)(void *arg, int direction), mus_float_t *filter, mus_long_t fftsize, mus_long_t filtersize, void *closure)

    mus_float_t *mus_spectrum(mus_float_t *rdat, mus_float_t *idat, mus_float_t *window, mus_long_t n, mus_spectrum_t type)
    void mus_fft(mus_float_t *rl, mus_float_t *im, mus_long_t n, int di)
    mus_float_t *mus_make_fft_window(mus_fft_window_t type, mus_long_t size, mus_float_t beta)
    mus_float_t *mus_make_fft_window_with_window(mus_fft_window_t type, mus_long_t size, mus_float_t beta, mus_float_t mu, mus_float_t *window)
    const char *mus_fft_window_name(mus_fft_window_t win)
    const char **mus_fft_window_names()

    mus_float_t *mus_autocorrelate(mus_float_t *data, mus_long_t n)
    mus_float_t *mus_correlate(mus_float_t *data1, mus_float_t *data2, mus_long_t n)
    mus_float_t *mus_convolution(mus_float_t *rl1, mus_float_t *rl2, mus_long_t n)
    void mus_convolve_files(const char *file1, const char *file2, mus_float_t maxamp, const char *output_file)
    mus_float_t *mus_cepstrum(mus_float_t *data, mus_long_t n)

    bint mus_is_granulate(mus_any *ptr)
    mus_float_t mus_granulate(mus_any *ptr, mus_float_t (*input)(void *arg, int direction))
    mus_float_t mus_granulate_with_editor(mus_any *ptr, mus_float_t (*input)(void *arg, int direction), int (*edit)(void *closure))
    mus_any *mus_make_granulate(mus_float_t (*input)(void *arg, int direction), 
                       mus_float_t expansion, mus_float_t length, mus_float_t scaler, 
                       mus_float_t hop, mus_float_t ramp, mus_float_t jitter, int max_size, 
                       int (*edit)(void *closure),
                       void *closure)
    int mus_granulate_grain_max_length(mus_any *ptr)
    void mus_granulate_set_edit_function(mus_any *ptr, int (*edit)(void *closure))

    mus_long_t mus_set_file_buffer_size(mus_long_t size)
    mus_long_t mus_file_buffer_size()

    mus_float_t mus_apply(mus_any *gen, mus_float_t f1, mus_float_t f2)

    bint mus_is_phase_vocoder(mus_any *ptr)
    mus_any *mus_make_phase_vocoder(mus_float_t (*input)(void *arg, int direction), 
                       int fftsize, int overlap, int interp,
                       mus_float_t pitch,
                       bint (*analyze)(void *arg, mus_float_t (*input)(void *arg1, int direction)),
                       int (*edit)(void *arg), #/* return value is ignored (int return type is intended to be consistent with granulate) */
                       mus_float_t (*synthesize)(void *arg), 
                       void *closure)
    mus_float_t mus_phase_vocoder(mus_any *ptr, mus_float_t (*input)(void *arg, int direction))
    mus_float_t mus_phase_vocoder_with_editors(mus_any *ptr, 
                        mus_float_t (*input)(void *arg, int direction),
                        bint (*analyze)(void *arg, mus_float_t (*input)(void *arg1, int direction)),
                        int (*edit)(void *arg), 
                        mus_float_t (*synthesize)(void *arg))

    mus_float_t *mus_phase_vocoder_amp_increments(mus_any *ptr)
    mus_float_t *mus_phase_vocoder_amps(mus_any *ptr)
    mus_float_t *mus_phase_vocoder_freqs(mus_any *ptr)
    mus_float_t *mus_phase_vocoder_phases(mus_any *ptr)
    mus_float_t *mus_phase_vocoder_phase_increments(mus_any *ptr)


    mus_any *mus_make_ssb_am(mus_float_t freq, int order)
    bint mus_is_ssb_am(mus_any *ptr)
    mus_float_t mus_ssb_am_unmodulated(mus_any *ptr, mus_float_t insig)
    mus_float_t mus_ssb_am(mus_any *ptr, mus_float_t insig, mus_float_t fm)

    void mus_clear_sinc_tables()
    void *mus_environ(mus_any *gen)
    void *mus_set_environ(mus_any *gen, void *e)
    mus_any *mus_bank_generator(mus_any *g, int i)
    

    
    







#==================================================================================
# The code is part of an attempt at translation of Bill Schottstedaet's sndlib 
# available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

cdef extern from "sndlib.h":
    ctypedef signed long  int64_t
    ctypedef double mus_float_t
    ctypedef unsigned char uint8_t
    
    ctypedef int64_t mus_long_t 
    
    
    ctypedef enum mus_header_t:
        pass

    ctypedef enum mus_sample_t:
        pass
        

    int mus_sound_initialize()
    ctypedef void mus_error_handler_t(int type, char *msg)
    mus_error_handler_t *mus_error_set_handler(mus_error_handler_t *new_error_handler)
    const char *mus_error_type_to_string(int err)

    ctypedef void mus_print_handler_t(char *msg)
    mus_print_handler_t *mus_print_set_handler(mus_print_handler_t *new_print_handler)

    ctypedef mus_float_t mus_clip_handler_t(mus_float_t val)
    #mus_clip_handler_t *mus_clip_set_handler(mus_clip_handler_t *new_clip_handler)
    #mus_clip_handler_t *mus_clip_set_handler_and_checker(mus_clip_handler_t *new_clip_handler, bool (*checker)(void))

    mus_long_t mus_sound_samples(const char *arg)
    mus_long_t mus_sound_framples(const char *arg)
    int mus_sound_datum_size(const char *arg)
    mus_long_t mus_sound_data_location(const char *arg)
    int mus_sound_chans(const char *arg)
    int mus_sound_srate(const char *arg)
    mus_header_t mus_sound_header_type(const char *arg)
    mus_sample_t mus_sound_sample_type(const char *arg)
    int mus_sound_original_sample_type(const char *arg)
   
    mus_long_t mus_sound_comment_start(const char *arg)
    mus_long_t mus_sound_comment_end(const char *arg)
    mus_long_t mus_sound_length(const char *arg)
    int mus_sound_fact_samples(const char *arg)
    #time_t mus_sound_write_date(const char *arg)
    int mus_sound_type_specifier(const char *arg)
    int mus_sound_block_align(const char *arg)
    int mus_sound_bits_per_sample(const char *arg)

    int mus_sound_set_chans(const char *arg, int val)
    int mus_sound_set_srate(const char *arg, int val)
    mus_header_t mus_sound_set_header_type(const char *arg, mus_header_t val)
    mus_sample_t mus_sound_set_sample_type(const char *arg, mus_sample_t val)
    int mus_sound_set_data_location(const char *arg, mus_long_t val)
    int mus_sound_set_samples(const char *arg, mus_long_t val)

    const char *mus_header_type_name(mus_header_t type)
    const char *mus_header_type_to_string(mus_header_t type)
    const char *mus_sample_type_name(mus_sample_t samp_type)
    const char *mus_sample_type_to_string(mus_sample_t samp_type)
    const char *mus_sample_type_short_name(mus_sample_t samp_type)

    char *mus_sound_comment(const char *name)
    int mus_bytes_per_sample(mus_sample_t samp_type)
    float mus_sound_duration(const char *arg)
    int mus_sound_initialize()
    int mus_sound_override_header(const char *arg, int srate, int chans, mus_sample_t samp_type, mus_header_t type, mus_long_t location, mus_long_t size)
    int mus_sound_forget(const char *name)
    int mus_sound_prune()
   # void mus_sound_report_cache(FILE *fp)
    int *mus_sound_loop_info(const char *arg)
    void mus_sound_set_loop_info(const char *arg, int *loop)
    int mus_sound_mark_info(const char *arg, int **mark_ids, int **mark_positions)

    int mus_sound_open_input(const char *arg)
    int mus_sound_open_output(const char *arg, int srate, int chans, mus_sample_t sample_type, mus_header_t header_type, const char *comment)
    int mus_sound_reopen_output(const char *arg, int chans, mus_sample_t samp_type, mus_header_t type, mus_long_t data_loc)
    int mus_sound_close_input(int fd)
    int mus_sound_close_output(int fd, mus_long_t bytes_of_data)
    #define mus_sound_read(Fd, Beg, End, Chans, Bufs) mus_file_read(Fd, Beg, End, Chans, Bufs)
    #define mus_sound_write(Fd, Beg, End, Chans, Bufs) mus_file_write(Fd, Beg, End, Chans, Bufs)

    mus_long_t mus_sound_maxamps(const char *ifile, int chans, mus_float_t *vals, mus_long_t *times)
    int mus_sound_set_maxamps(const char *ifile, int chans, mus_float_t *vals, mus_long_t *times)
    bint mus_sound_maxamp_exists(const char *ifile)
    bint mus_sound_channel_maxamp_exists(const char *file, int chan)
    mus_float_t mus_sound_channel_maxamp(const char *file, int chan, mus_long_t *pos)
    void mus_sound_channel_set_maxamp(const char *file, int chan, mus_float_t mx, mus_long_t pos)
    mus_long_t mus_file_to_array(const char *filename, int chan, mus_long_t start, mus_long_t samples, mus_float_t *array)
    int mus_array_to_file(const char *filename, mus_float_t *ddata, mus_long_t len, int srate, int channels)
    const char *mus_array_to_file_with_error(const char *filename, mus_float_t *ddata, mus_long_t len, int srate, int channels)
    mus_long_t mus_file_to_float_array(const char *filename, int chan, mus_long_t start, mus_long_t samples, mus_float_t *array)
    int mus_float_array_to_file(const char *filename, mus_float_t *ddata, mus_long_t len, int srate, int channels)

    mus_float_t **mus_sound_saved_data(const char *arg)
    void mus_sound_set_saved_data(const char *arg, mus_float_t **data)
    void mus_file_save_data(int tfd, mus_long_t framples, mus_float_t **data)

    int mus_audio_open_output(int dev, int srate, int chans, mus_sample_t samp_type, int size)
    int mus_audio_open_input(int dev, int srate, int chans, mus_sample_t samp_type, int size)
    int mus_audio_write(int line, char *buf, int bytes)
    int mus_audio_close(int line)
    int mus_audio_read(int line, char *buf, int bytes)

    int mus_audio_initialize()
    char *mus_audio_moniker()
    int mus_audio_api()
    mus_sample_t mus_audio_compatible_sample_type(int dev)

    #if HAVE_OSS || HAVE_ALSA
    void mus_oss_set_buffers(int num, int size)
    char *mus_alsa_playback_device()
    char *mus_alsa_set_playback_device(const char *name)
    char *mus_alsa_capture_device()
    char *mus_alsa_set_capture_device(const char *name)
    char *mus_alsa_device()
    char *mus_alsa_set_device(const char *name)
    int mus_alsa_buffer_size()
    int mus_alsa_set_buffer_size(int size)
    int mus_alsa_buffers()
    int mus_alsa_set_buffers(int num)
    bint mus_alsa_squelch_warning()
    bint mus_alsa_set_squelch_warning(bint val)
    #endif

    #if __APPLE__
    bint mus_audio_output_properties_mutable(bint mut)
    #endif

    int mus_audio_device_channels(int dev)
    mus_sample_t mus_audio_device_sample_type(int dev)


    int mus_file_open_descriptors(int tfd, const char *arg, mus_sample_t df, int ds, mus_long_t dl, int dc, mus_header_t dt)
    int mus_file_open_read(const char *arg)
    bint mus_file_probe(const char *arg)
    int mus_file_open_write(const char *arg)
    int mus_file_create(const char *arg)
    int mus_file_reopen_write(const char *arg)
    int mus_file_close(int fd)
    mus_long_t mus_file_seek_frample(int tfd, mus_long_t frample)
    mus_long_t mus_file_read(int fd, mus_long_t beg, mus_long_t end, int chans, mus_float_t **bufs)
    mus_long_t mus_file_read_chans(int fd, mus_long_t beg, mus_long_t end, int chans, mus_float_t **bufs, mus_float_t **cm)
    int mus_file_write(int tfd, mus_long_t beg, mus_long_t end, int chans, mus_float_t **bufs)
    mus_long_t mus_file_read_any(int tfd, mus_long_t beg, int chans, mus_long_t nints, mus_float_t **bufs, mus_float_t **cm)
    mus_long_t mus_file_read_file(int tfd, mus_long_t beg, int chans, mus_long_t nints, mus_float_t **bufs)
    mus_long_t mus_file_read_buffer(int charbuf_sample_type, mus_long_t beg, int chans, mus_long_t nints, mus_float_t **bufs, char *charbuf)
    int mus_file_write_file(int tfd, mus_long_t beg, mus_long_t end, int chans, mus_float_t **bufs)
    int mus_file_write_buffer(int charbuf_sample_type, mus_long_t beg, mus_long_t end, int chans, mus_float_t **bufs, char *charbuf, bint clipped)
    char *mus_expand_filename(const char *name)
    char *mus_getcwd()

    bint mus_clipping()
    bint mus_set_clipping(bint new_value)
    bint mus_file_clipping(int tfd)
    int mus_file_set_clipping(int tfd, bint clipped)

    int mus_file_set_header_type(int tfd, mus_header_t type)
    mus_header_t mus_file_header_type(int tfd)
    char *mus_file_fd_name(int tfd)
    int mus_file_set_chans(int tfd, int chans)

    int mus_iclamp(int lo, int val, int hi)
    mus_long_t mus_oclamp(mus_long_t lo, mus_long_t val, mus_long_t hi)
    mus_float_t mus_fclamp(mus_float_t lo, mus_float_t val, mus_float_t hi)

    void mus_reset_io_c()
    void mus_reset_headers_c()
    void mus_reset_audio_c()

    int mus_samples_bounds(uint8_t *data, int bytes, int chan, int chans, mus_sample_t samp_type, mus_float_t *min_samp, mus_float_t *max_samp)

    mus_long_t mus_max_malloc()
    mus_long_t mus_set_max_malloc(mus_long_t new_max)
    mus_long_t mus_max_table_size()
    mus_long_t mus_set_max_table_size(mus_long_t new_max)

    char *mus_strdup(const char *str)
    int mus_strlen(const char *str)
    bint mus_strcmp(const char *str1, const char *str2)
    char *mus_strcat(char *errmsg, const char *str, int *err_size)


    bint mus_is_sample_type(int n)
    bint mus_is_header_type(int n)


    mus_long_t mus_header_samples()
    mus_long_t mus_header_data_location()
    int mus_header_chans()
    int mus_header_srate()
    mus_header_t mus_header_type()
    mus_sample_t mus_header_sample_type()
    mus_long_t mus_header_comment_start()
    mus_long_t mus_header_comment_end()
    int mus_header_type_specifier()
    int mus_header_bits_per_sample()
    int mus_header_fact_samples()
    int mus_header_block_align()
    int mus_header_loop_mode(int which)
    int mus_header_loop_start(int which)
    int mus_header_loop_end(int which)
    int mus_header_mark_position(int id)
    int mus_header_mark_info(int **marker_ids, int **marker_positions)
    int mus_header_base_note()
    int mus_header_base_detune()
    void mus_header_set_raw_defaults(int sr, int chn, mus_sample_t frm)
    void mus_header_raw_defaults(int *sr, int *chn, mus_sample_t *frm)
    mus_long_t mus_header_true_length()
    int mus_header_original_sample_type()
    mus_long_t mus_samples_to_bytes(mus_sample_t samp_type, mus_long_t size)
    mus_long_t mus_bytes_to_samples(mus_sample_t samp_type, mus_long_t size)
    int mus_header_read(const char *name)
    int mus_header_write(const char *name, mus_header_t type, int srate, int chans, mus_long_t loc, mus_long_t size_in_samples, 
        mus_sample_t samp_type, const char *comment, int len)
    int mus_write_header(const char *name, mus_header_t type, int in_srate, int in_chans, mus_long_t size_in_samples, 
        mus_sample_t samp_type, const char *comment)
    mus_long_t mus_header_aux_comment_start(int n)
    mus_long_t mus_header_aux_comment_end(int n)
    int mus_header_initialize()
    bint mus_header_writable(mus_header_t type, mus_sample_t samp_type)
    void mus_header_set_aiff_loop_info(int *data)
    int mus_header_sf2_entries()
    char *mus_header_sf2_name(int n)
    int mus_header_sf2_start(int n)
    int mus_header_sf2_end(int n)
    int mus_header_sf2_loop_start(int n)
    int mus_header_sf2_loop_end(int n)
    const char *mus_header_original_sample_type_name(int samp_type, mus_header_t type)
    bint mus_header_no_header(const char *name)

    char *mus_header_riff_aux_comment(const char *name, mus_long_t *starts, mus_long_t *ends)
    char *mus_header_aiff_aux_comment(const char *name, mus_long_t *starts, mus_long_t *ends)

    int mus_header_change_chans(const char *filename, mus_header_t type, int new_chans)
    int mus_header_change_srate(const char *filename, mus_header_t type, int new_srate)
    int mus_header_change_type(const char *filename, mus_header_t new_type, mus_sample_t new_sample_type)
    int mus_header_change_sample_type(const char *filename, mus_header_t type, mus_sample_t new_sample_type)
    int mus_header_change_location(const char *filename, mus_header_t type, mus_long_t new_location)
    int mus_header_change_comment(const char *filename, mus_header_t type, const char *new_comment)
    int mus_header_change_data_size(const char *filename, mus_header_t type, mus_long_t bytes)
    
    #complex double mus_edot_product(complex double freq, complex double *data, mus_long_t size)

    #typedef void mus_header_write_hook_t(const char *filename)
    #mus_header_write_hook_t *mus_header_write_set_hook(mus_header_write_hook_t *new_hook)
    
    
	



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses nrssb generator


# --------------- oboish ---------------- #
@cython.ccall
def oboish(beg, dur, freq, amp, aenv):
    res1: cython.double = max(1, round(1400.0 / max(1.0, min(1400.0, freq))))
    mod1 = clm.make_oscil(5.)
    res2: cython.double = max(1, round(2400.0 / max(1.0, min(2400.0, freq))))
    gen3 = clm.make_oscil(freq)
    start: cython.long = clm.seconds2samples(beg)
    amplitude = clm.make_env(aenv, duration=dur, base=4, scaler=amp)
    skenv = clm.make_env([0.0,0.0,1,1,2.0,clm.random(1.), 3.0, 0.0, max(4.0, dur*20.0), 0.0], duration=dur, scaler=clm.hz2radians(random.uniform(0.,freq*.05)))
    relamp: cython.double = .85 + random.uniform(0, .1)
    avib = clm.make_rand_interp(5, .2)
    hfreq = clm.hz2radians(freq)
    h3freq = clm.hz2radians(.003 * freq)
    
    scl: cython.double = .05 / amp
    gen = gens.make_nrssb(frequency=freq*res1, ratio=1/res1, n=res1, r=.75)
    gen2 = clm.make_oscil(freq*res2)
    stop: cython.long = start + clm.seconds2samples(dur)
    i: cython.long = 0
    for i in range(start, stop):
        vol: cython.double = (.8 + clm.rand_interp(avib)) * clm.env(amplitude)
        vola: cython.double = scl * vol
        vib: cython.double = (h3freq * clm.oscil(mod1)) + clm.env(skenv)
        result: cython.double = vol * (((relamp - vola) * gens.nrssb_interp(gen, (res1*vib), -1.0)) +
                         (((1 + vola) - relamp)  * clm.oscil(gen2, ((vib*res2)+(hfreq*clm.oscil(gen3, vib))))))
        clm.outa(i, result)
        if clm.default.reverb :
            clm.outa(i, (.01*result), clm.default.reverb)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm


# --------------- spectra ---------------- #
@cython.ccall
def spectra(startime, duration, frequency, amplitude, partials=[1,1,2,0.5], amp_envelope=[0,0,50,1,100,0], 
                vibrato_amplitude=0.005, vibrato_speed=5.0,degree=0.0, distance=1.0, reverb_amount=0.005):

    beg: cython.long  = clm.seconds2samples(startime)
    end:  cython.long = clm.seconds2samples(startime + duration)
    waveform: np.ndarray = clm.partials2wave(partials)
    freq: cython.double = clm.hz2radians(frequency)
    s = clm.make_table_lookup(frequency=frequency, wave=waveform)
    amp_env = clm.make_env(amp_envelope, scaler=amplitude, duration=duration)
    per_vib = clm.make_triangle_wave(vibrato_speed, vibrato_amplitude * freq)
    loc = clm.make_locsig(degree, distance, reverb_amount)
    ran_vib = clm.make_rand_interp(vibrato_speed + 1.0, vibrato_amplitude * freq)
    
    i: cython.long = beg
    
    for i in range(beg, end):
        clm.locsig(loc, i, clm.env(amp_env) * clm.table_lookup(s, clm.triangle_wave(per_vib) + clm.rand_interp(ran_vib)))

if __name__ == '__main__': 
    with clm.Sound( play = True, statistics=True):
        spectra(0, 4, 440.0, .1, [1.0,.4,2.0,.2,3.0,.2,4.0,.1,6.0,.1], [0.0,0.0,1.0,1.0,5.0,0.9,12.0,0.5,25.0,0.25,100.0,0.0])



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm


# --------------- metal ---------------- #
#   ;; from Perry Cook's HeavyMtl.cpp
@cython.ccall
def metal(beg, dur, freq, amp):
    osc0 = clm.make_oscil(freq)
    osc1 = clm.make_oscil(freq * 4.0 * 0.999)
    osc2 = clm.make_oscil(freq * 3.0 * 1.001)
    osc3 = clm.make_oscil(freq * 0.50 * 1.002)
    ampenv0 = clm.make_env([0,0,.001,1, (dur-.002),1,dur,0], duration=dur, scaler=(amp * .615))
    ampenv1 = clm.make_env([0,0,.001,1,(dur-.011),1,dur,0], duration=dur, scaler=.202)
    ampenv2 = clm.make_env([0,0,.01,1,(dur-.015),1,dur,0], duration=dur, scaler=.574)
    ampenv3 = clm.make_env([0,0,.03,1,(dur-.040),1,dur,0], duration=dur, scaler=.116)
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    
    i: cython.long = 0
    
    for i in range(st, nd):
        clm.outa(i, (clm.env(ampenv0) * (clm.oscil(osc0, ((clm.env(ampenv1) * clm.oscil(osc1, clm.env(ampenv2) * clm.oscil(osc2))))) +
                    (clm.env(ampenv3) * clm.oscil(osc3)))))



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm

if cython.compiled:
    from cython.cimports.pysndlib import clm
# from pysndlib.env import stretch_envelope
# --------------- pluck ---------------- #

# ;;; The Karplus-Strong algorithm as extended by David Jaffe and Julius Smith -- see 
# ;;;  Jaffe and Smith, "Extensions of the Karplus-Strong Plucked-String Algorithm"
# ;;;  CMJ vol 7 no 2 Summer 1983, reprinted in "The Music Machine".
# ;;;  translated from CLM's pluck.ins


def tuneIt(f, s1):
    def getOptimumC(S,o,p):
        pa = (1.0 / o) * math.atan2((S * math.sin(o)), (1.0 + (S * math.cos(o))) - S)
        tmpInt = math.floor(p - pa)
        pc = p - pa - tmpInt
        
        if pc < .1:
            while pc >= .1:
                tmpInt = tmpInt - 1
                pc += 1.0
        return [tmpInt, (math.sin(o) - math.sin(o * pc)) / math.sin(o + (o * pc))]
        
    p = clm.get_srate() / f
    s = .5 if s1 == 0.0 else s1
    o = clm.hz2radians(f)
    vals = getOptimumC(s,o,p)
    vals1 = getOptimumC((1.0 - s), o, p)
    
    if s != 1/2 and math.fabs(vals[1]) < math.fabs(vals1[1]):
        return [(1.0 - s), vals[1], vals[0]]
    else:
        return [s, vals1[1], vals1[0]]


@cython.ccall
def pluck(start, dur, freq, amp, weighting=.5, lossfact=.9):
    vals = tuneIt(freq, weighting)
    wt0 = vals[0]
    c = vals[1]
    dlen = vals[2]
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start + dur)
    lf = 1.0 if (lossfact == 0.0) else min(1.0, lossfact)
    wt = .5 if (wt0 == 0.0) else min(1.0, wt0)
    tab: np.ndarray  = np.zeros(dlen, dtype=np.double)
    allp = clm.make_one_zero(lf * (1.0 - wt), (lf * wt))
    feedb = clm.make_one_zero(c, 1.0)
    c1: cython.double = 1.0 - c
    
    k: cython.long = 0
    
    for k in range(0,dlen):
        tab[k] = clm.random(1.0) # could replace with python random
        
    ctr: cython.long = 0

    i: cython.long = 0
    val: cython.double = 0.
    
    for i in range(beg, end):
        ctr = (ctr + 1) % dlen
        val = (c1 * (clm.one_zero(feedb, clm.one_zero(allp, tab[ctr]))))
        tab[ctr] = val
        clm.outa(i, amp * val)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses r2k!cos generator

# --------------- pianoy2 ---------------- #
@cython.ccall
def pianoy2(start, dur, freq, amp, bounce_freq=5, bounce_amp=20):
    gen = gens.make_r2kfcos(freq, r=.5, k=3)
    ampf = clm.make_env([0,0,.01,1,.03,1,1,.15, max(2,dur), 0.0], base=32, scaler=amp, duration=dur)
    knock = gens.make_fmssb(10., 20., index=1.0)
    kmpf = clm.make_env([0,0,1,1,3,1,100,0], base=3, scaler=.05, length=30000)
    indf = clm.make_env([0,1,1,0], length=30000, base=3, scaler=10)
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start+dur)
    
    for i in range(beg, end):
        knock.index = clm.env(indf)
        clm.outa(i, (clm.env(ampf) * gens.r2kfcos(gen)) + (clm.env(kmpf) * gens.fmssb(knock, 0.0)))

    
if __name__ == '__main__': 
    with clm.Sound(statistics=True, play=True):
        pianoy2(0,1,100,.5)
